package hSPEA2;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

// SPEA2 Implemented according to Zitzler, E., Laumanns, M., & Thiele, L. (2001). SPEA2: Improving the strength Pareto evolutionary algorithm.
public class SPEA2 {
	
	int population_size;
	int archive_size;
	int simulation_runs;
	double scheduling_confidence;
	double T;
	int k;
	int max_generations;
	double winner_selection_probability;
	double mutation_rate;
	double crossover_rate;
	ArrayList<Individual> population;
	ArrayList<Individual> archive;
	
	FileReader fr;
	LocalSearch ls;
	
	public SPEA2(int pop_size, int gen, int sim_runs, double sched_conf, double t){
		//Set from experimenter
		population_size = pop_size;
		archive_size = pop_size; //Should be tested
		max_generations = gen;
		simulation_runs = sim_runs;
		scheduling_confidence = sched_conf;
		T=t;
		
		//Set internally
		k = (int) Math.round(Math.sqrt(population_size+archive_size)); //Common setting for SPEA2
		winner_selection_probability = 1;
		mutation_rate = 0.2;
		mutation_rate = 0.9;
		
		//Initialization
		population = new ArrayList<Individual>();
		archive = new ArrayList<Individual>();
	}
	
	public ArrayList<Individual> main_loop(String instance, double std){
		PerformanceManager PM = new PerformanceManager();
		fr = new FileReader(instance, std);
		fr.sample_durations(simulation_runs);
		//ls = new LocalSearch(fr);
		int generation = 0;
		initialization();
		while (true){
			mutate_population();
			fitness_assignment();
			environmental_selection();
			if(generation >= max_generations+1){
				return archive; // TERMINATION
			}
			mating_selection();
			generation++;	
			fr.sample_durations(simulation_runs);
		}
	}
	
	//INITIALIZATION
	public void initialization(){
		while (population.size() < population_size){
			population.add(new Individual(fr, population_size, archive_size,simulation_runs, scheduling_confidence, T));
		}
	}
	
	//LOCAL SEARCH
	public void local_search(int fitness_index){
		ArrayList<Individual> improved_population = new ArrayList<Individual>();
		for(Individual i : population){
			improved_population.add(ls.random_neighbourhood_search(i));
		}
		population = improved_population;
	}
	
	//FITNESS ASSIGNMENT
	public void fitness_assignment(){
			assign_strength_values();
			assign_raw_fitness();
			assign_distances_and_density();
			for(Individual i : population){
				i.set_fitness(i.raw_fitness+i.density);
			}
			for(Individual i : archive){
				i.set_fitness(i.raw_fitness+i.density);
			}
	}
	
	public void assign_strength_values(){
		for(Individual i : population){
			int strength = 0;
			for(Individual j : population){
				if(i.dominates(j)){
					strength++;
				}
			}
			for(Individual j : archive){
				if(i.dominates(j)){
					strength++;
				}
			}
			i.set_strength_value(strength);
		}
		for(Individual i : archive){
			int strength = 0;
			for(Individual j : population){
				if(i.dominates(j)){
					strength++;
				}
			}
			for(Individual j : archive){
				if(i.dominates(j)){
					strength++;
				}
			}
			i.set_strength_value(strength);
		}
	}
	
	public void assign_raw_fitness(){
		for(Individual i : population){
			int raw_fitness = 0;
			for(Individual j : population){
				if(j.dominates(i)){
					raw_fitness+=j.get_strength_value();
				}
			}
			for(Individual j : archive){
				if(j.dominates(i)){
					raw_fitness+=j.get_strength_value();
				}
			}
			i.set_raw_fitness(raw_fitness);
		}
		for(Individual i : archive){
			int raw_fitness = 0;
			for(Individual j : population){
				if(j.dominates(i)){
					raw_fitness+=j.get_strength_value();
				}
			}
			for(Individual j : archive){
				if(j.dominates(i)){
					raw_fitness+=j.get_strength_value();
				}
			}
			i.set_raw_fitness(raw_fitness);
		}
	}
	
	public void assign_distances_and_density(){
		for(Individual i : population){
			int l = 0;
			for(Individual j : population){
				i.set_distance_to_neighbour(i.equlidean_distance(j), l);
				l++;
			}
			for(Individual j : archive){
				i.set_distance_to_neighbour(i.equlidean_distance(j), l);
				l++;
			}
			i.sort_distances();
			i.set_density(1/(i.distances[k]+2));
		}
		for(Individual i : archive){
			int l = 0;
			for(Individual j : population){
				i.set_distance_to_neighbour(i.equlidean_distance(j), l);
				l++;
			}
			for(Individual j : archive){
				i.set_distance_to_neighbour(i.equlidean_distance(j), l);
				l++;
			}
			i.sort_distances();
			i.set_density(1/(i.distances[k]+2));
		}
	}
	
	//ENVIRONMENTAL SELECTION
	public void environmental_selection(){
		//Copy all nondominated individuals from the population and the archive into a new archive
		ArrayList<Individual> next_generation_archive = new ArrayList<Individual>();
		for(Individual i : population){
			if(i.get_fitness()<1){ //Nondominated
				next_generation_archive.add(i);
			}
		}
		for(Individual i : archive){
			if(i.get_fitness()<1){ //Nondominated
				next_generation_archive.add(i);
			}
		}
		archive = next_generation_archive; // Update archive
		remove_identical_individuals_in_archive();
		//Three scenarios
		if(archive.size() < archive_size){
			//Merge population and archive, sort and fill next generation archive until full
			List<Individual> population_and_archive = new ArrayList<Individual>();
			for(Individual i : population){
				if(i.get_fitness()>1){
					population_and_archive.add(i);
				}
			}
			for(Individual i : archive){
				if(i.get_fitness()>=1){
					population_and_archive.add(i);
				}
			}
			Collections.sort(population_and_archive, Individual.COMPARE_BY_FITNESS);
			
			while(archive.size() < archive_size){
				if(population_and_archive.size()>0){
					archive.add(population_and_archive.get(population_and_archive.size()-1));
					population_and_archive.remove(population_and_archive.size()-1);
				}
				else{
					archive.add(new Individual(fr, population_size, population_size+archive_size, simulation_runs, scheduling_confidence, T));
				}
			}
		}
		else if(archive.size() > archive_size){
			while(archive.size() > archive_size){
				double minimum_distance = Double.POSITIVE_INFINITY; 
				int index_of_minimum_distance = 0;
				for (int i=0; i<archive.size(); i++){
					Individual subject = archive.get(i);
					for(Individual j : archive){
						double distance = subject.equlidean_distance(j);
						if (distance < minimum_distance && subject.equlidean_distance(j)!=0){
							minimum_distance = distance;
							index_of_minimum_distance = i;
						}
					}
				}
				archive.remove(index_of_minimum_distance);
			}
		}
		//Else: the next generation archive size is exactly equal to the archive size
		population.clear(); //Clear the current population of all individuals
	}
	
	//MATING SELECTION
	public void mating_selection(){
		fitness_assignment();
		while(population.size() < population_size){
			population.add(crossover(binary_tournament_selection(), binary_tournament_selection())); 
		}
	}
	
	public Individual binary_tournament_selection(){
		int i = (int) (Math.random()*(archive_size-0.5));
		int j = (int) (Math.random()*(archive_size-0.5));
		return binary_tournament(archive.get(i), archive.get(j));
	}
	
	public Individual binary_tournament(Individual i, Individual j){
		double r = Math.random();
		if(i.fitness < j.fitness && r < winner_selection_probability){
			return i;
		}
		else return j; //Random
	}
	
	public Individual crossover(Individual p1, Individual p2){
		if(Math.random()>crossover_rate){
			return p1;
		}
		else{
		Individual parent1 = p1.clone();
		Individual parent2 = p2.clone();
		int lower_index = (int)(Math.random()*(fr.N-0.5)-1);
		int upper_index = lower_index+(int)(Math.random()*(fr.N-lower_index-0.5));
		ArrayList<Integer> parent2_priority = new ArrayList<Integer>();
		for(int i =0;i<parent2.get_chromosome().size();i++){
			parent2_priority.add(parent2.get_chromosome().get(i).getPriority());
		}
		Individual child = new Individual(fr, population_size, archive_size, simulation_runs, scheduling_confidence, T);
		for(int i=lower_index;i<upper_index;i++){
			child.get_chromosome().set(i,parent1.get_chromosome().get(i));
			for(int j=0;j<parent2_priority.size();j++){
				if (parent1.get_chromosome().get(i).getPriority()==parent2_priority.get(j)){
					parent2_priority.remove(j);
				}
			}
		}
		for(int i=0;i<lower_index;i++){
			child.get_chromosome().set(i,parent2.get_chromosome().get(i));
			child.get_chromosome().get(i).setPriority(parent2_priority.get(i));
		}
		for(int i=upper_index;i<fr.N;i++){
			child.get_chromosome().set(i,parent2.get_chromosome().get(i));
			child.get_chromosome().get(i).setPriority(parent2_priority.get(i-(upper_index-lower_index)));
		}
		
		child.set_fitness(child.get_chromosome());
		return child;
		}
	}
	
	//VARIATION
	public void mutate_population(){
		ArrayList<Individual> mutated_population = new ArrayList<Individual>();
		for (int i =0; i< population.size(); i++){
			if(Math.random() < mutation_rate){
				Individual mutant = mutate(population.get(i));
				mutant.set_fitness(mutant.get_chromosome());
				mutated_population.add(mutant);
			}
			else{
				mutated_population.add(population.get(i));
			}
		}
		population = mutated_population;
	}
	
	public void replace_identical_individuals(){
		for(int i=0;i<population.size();i++){
			for(int j=i+1;j<population.size();j++){
				if(population.get(i).equal_to(population.get(j), 1)){
					population.remove(j);
					population.add(new Individual(fr, population_size, population_size+archive_size, simulation_runs, scheduling_confidence, T));
				}
			}
		}
	}
	
	public void remove_identical_individuals_in_archive(){
		for(int l=0;l<archive.size();l++){
			Individual i = archive.get(l);
			for(int k=l+1;k<archive.size();k++){
				Individual j = archive.get(k);
				if(!i.equals(j) && i.equlidean_distance(j)==0){
					archive.remove(j);
				}
			}
		}
	}
	
	public Individual mutate(Individual individual){
		double r = Math.random();
		if(r<0.33){
			return switchRoomDayMutation(individual);
		}
		else if(r<0.66){
			return switchPriorityMutation(individual);
		}
		else return newRoomDayMutation(individual);
	}
	
	public Individual switchPriorityMutation(Individual current){
		int i = (int)(Math.random()*current.get_chromosome().size());
		int j = (int)(Math.random()*current.get_chromosome().size());
		if(fr.specialty[i]==fr.specialty[j] && i!=j){
			int priority1 = current.get_chromosome().get(i).getPriority();
			int priority2 = current.get_chromosome().get(j).getPriority();
			if(priority1 != priority2){
				Individual n = current.clone();
				n.get_chromosome().get(i).setPriority(priority2);
				n.get_chromosome().get(j).setPriority(priority1);
				n.set_fitness(n.get_chromosome());
				current = n;
			}
		}
		return current;
	}
	
	public Individual switchRoomDayMutation(Individual current){
		int i = (int)(Math.random()*current.get_chromosome().size());
		int j = (int)(Math.random()*current.get_chromosome().size());
		if(fr.specialty[i]==fr.specialty[j] && i!=j){
			int day1 = current.get_chromosome().get(i).getDay();
			int day2 = current.get_chromosome().get(j).getDay();
			int room1 = current.get_chromosome().get(i).getRoom();
			int room2 = current.get_chromosome().get(j).getRoom();
			if(day1 != day2){
				Individual n = current.clone();
				n.get_chromosome().get(i).setDay(day2);
				n.get_chromosome().get(j).setDay(day1);
				n.get_chromosome().get(i).setRoom(room2);
				n.get_chromosome().get(j).setRoom(room1);
				n.set_fitness(n.get_chromosome());
				current = n;
			}
		}
		return current;
	}
	
	public Individual newRoomDayMutation(Individual current){
		int i = (int)(Math.random()*current.get_chromosome().size());
		int s = fr.specialty[i];
		double p = Math.random();
		for(int r=0;r<fr.R;r++){
			for(int d=0;d<fr.D;d++){
				if(fr.MSS[r][d]==s && p<fr.MSS_probability[r][d]){
					Individual n = current.clone();
					n.get_chromosome().get(i).setDay(d);
					n.get_chromosome().get(i).setRoom(r);
					n.set_fitness(n.chromosome);
					return n;
				}
			}
		}
		return current;
	}
}

